// Application class implementation
#include "pch.h"
#include "Application.h"
#include "resources.h"
#include "utility.h"

// Define timer IDs
#define IDT_UPDATEAPP 1
#define IDT_SAVEDTEXT 2

// Define window message to update the timer display
#define WM_UPDATEAPP WM_USER + 1

// Enable visual styles
#pragma comment(linker, "\"/manifestdependency:type='win32' \
name='Microsoft.Windows.Common-Controls' version='6.0.0.0' \
processorArchitecture='*' publicKeyToken='6595b64144ccf1df' language='*'\"")

namespace mainspring {
    Application::Application(const char* appName, const char* appVersion)
        : m_appName(appName)
        , m_appVersion(appVersion)
        , m_dlg()
        , m_timing(false)
        , m_startpauseButtonPressed(false)
        , m_elapsedSaved()
        , m_startTime()
        , m_file()
        , m_showSavedText(false)
    {
        std::cout << std::format("{} {}\n", m_appName, m_appVersion);
    }

    void Application::run() {
        std::string configPath = getConfigPath();
        std::string recentFile;
        if (readFile(configPath.c_str(), recentFile) && !recentFile.empty())
            open(recentFile);

        m_dlg = CreateDialogParam(nullptr, MAKEINTRESOURCE(IDD_DIALOGMAIN), nullptr, reinterpret_cast<DLGPROC>(mainDlgProc), reinterpret_cast<LPARAM>(this));

        HACCEL hAccel = LoadAccelerators(GetModuleHandle(NULL), MAKEINTRESOURCE(IDA_ACCELMAIN));
        MSG msg = {};
        ShowWindow(m_dlg, SW_SHOW);
        while (GetMessage(&msg, nullptr, 0, 0))
            if (!TranslateAccelerator(m_dlg, hAccel, &msg))
                IsDialogMessage(m_dlg, &msg);
    }

    BOOL CALLBACK Application::mainDlgProc(HWND dlg, UINT msg, WPARAM wParam, LPARAM lParam) {
        static Application* app = nullptr;

        switch (msg) {
            case WM_INITDIALOG: {
                app = reinterpret_cast<Application*>(lParam);

                SetWindowText(dlg, app->m_appName);
                SendMessage(dlg, WM_SETICON, ICON_BIG, reinterpret_cast<LPARAM>(LoadIcon(GetModuleHandle(NULL), MAKEINTRESOURCE(IDI_ICONMAIN))));

                POINT cursor = {};
                GetCursorPos(&cursor);
                MONITORINFO mi = {};
                mi.cbSize = sizeof(mi);
                GetMonitorInfo(MonitorFromPoint(cursor, MONITOR_DEFAULTTONEAREST), &mi);
                SetWindowPos(dlg, nullptr, mi.rcMonitor.left + 100, mi.rcMonitor.top + 100, 0, 0, SWP_NOZORDER | SWP_NOSIZE);

                LOGFONT lFont = {};
                lFont.lfHeight = 30;
                HFONT titleFont = CreateFontIndirect(&lFont);
                SendMessage(GetDlgItem(dlg, IDC_STATICTITLE), WM_SETFONT, reinterpret_cast<WPARAM>(titleFont), NULL);

                std::memcpy(lFont.lfFaceName, "Consolas", 9);
                lFont.lfHeight = 60;
                HFONT timeFont = CreateFontIndirect(&lFont);
                SendMessage(GetDlgItem(dlg, IDC_STATICTIME), WM_SETFONT, reinterpret_cast<WPARAM>(timeFont), NULL);

                SendMessage(dlg, WM_UPDATEAPP, NULL, NULL);
                SetTimer(dlg, IDT_UPDATEAPP, 50, nullptr);

                return TRUE;
            }

            case WM_COMMAND:
                switch (LOWORD(wParam)) {
                    case IDC_BUTTONSTARTPAUSE:
                        app->m_startpauseButtonPressed = true;
                        return TRUE;

                    case ID_FILE_RESET: {
                        if (!app->m_file.empty())
                            app->save();

                        app->reset();
                        return TRUE;
                    }

                    case ID_FILE_SAVE: {
                        app->save();
                        return TRUE;
                    }

                    case ID_FILE_SAVEAS: {
                        app->save(true);
                        return TRUE;
                    }

                    case ID_FILE_OPEN: {
                        if (!app->m_file.empty())
                            app->save();

                        app->open();
                        return TRUE;
                    }

                    case ID_FILE_EXIT: {
                        SendMessage(dlg, WM_CLOSE, NULL, NULL);
                        return TRUE;
                    }

                    case ID_HELP_ABOUT:
                        MessageBox(dlg, std::format("{} {}\nCopyright Grayson Riffe 2025\ngraysonriffe.com\n\n"
                            "\tMainspring is a simple timekeeper. It can be used as a simple stopwatch for short-term activities, or "
                            "it can log hours on a longer-term project. Times can be saved to a file, and any subsequent uses will autosave to that file. The most recent "
                            "time will be loaded when Mainspring starts.", app->m_appName, app->m_appVersion).c_str(), std::format("About {}", app->m_appName).c_str(), MB_OK);
                        return TRUE;
                }

                return FALSE;

            case WM_TIMER:
                switch (wParam) {
                    case IDT_UPDATEAPP:
                        SendMessage(dlg, WM_UPDATEAPP, NULL, NULL);
                        return TRUE;

                    case IDT_SAVEDTEXT:
                        app->m_showSavedText = false;
                        return TRUE;
                }

                return FALSE;

                return TRUE;

            case WM_UPDATEAPP: {
                if (app->m_startpauseButtonPressed) {
                    app->m_startpauseButtonPressed = false;

                    if (!app->m_timing)
                        app->m_startTime = Clock::now();
                    else
                        app->m_elapsedSaved += app->getElapsed();

                    app->m_timing = !app->m_timing;
                }

                if (app->m_showSavedText)
                    SetDlgItemText(dlg, IDC_STATICTITLE, "Saved!");

                else if (!app->m_file.empty()) {
                    std::string filename = app->m_file.substr(app->m_file.find_last_of('\\') + 1);
                    SetDlgItemText(dlg, IDC_STATICTITLE, filename.c_str());
                }
                else
                    SetDlgItemText(dlg, IDC_STATICTITLE, app->m_appName);

                if (app->m_timing)
                    SetDlgItemText(dlg, IDC_BUTTONSTARTPAUSE, "Pause");
                else
                    SetDlgItemText(dlg, IDC_BUTTONSTARTPAUSE, !app->m_elapsedSaved ? "Start" : "Resume");

                Seconds totalTime = app->m_elapsedSaved + app->getElapsed();

                uint32_t hr = totalTime / (3600), min = totalTime % (3600) / 60, sec = totalTime % 60;
                SetDlgItemText(dlg, IDC_STATICTIME, std::format("{:03}:{:02}:{:02}", hr, min, sec).c_str());

                return TRUE;
            }

            case WM_CLOSE: {
                if (!app->m_file.empty())
                    app->save();

                std::string configPath = app->getConfigPath();
                writeFile(configPath.c_str(), app->m_file);

                KillTimer(dlg, 1);
                DestroyWindow(dlg);
                return TRUE;
            }

            case WM_DESTROY:
                PostQuitMessage(0);
                return TRUE;
        }

        return FALSE;
    }

    std::string Application::getConfigPath() {
        char appDataPath[MAX_PATH] = {};
        SHGetFolderPath(nullptr, CSIDL_APPDATA, nullptr, NULL, appDataPath);
        return std::string(appDataPath) + "\\" + m_appName + "\\last";
    }

    Seconds Application::getElapsed() {
        if (m_timing)
            return Duration(Clock::now() - m_startTime).count();
        else
            return 0;
    }

    void Application::reset() {
        m_timing = false;
        m_elapsedSaved = 0;
        m_file.clear();
    }

    void Application::save(bool saveas) {
        if (m_file.empty() || saveas) {
            OPENFILENAME ofn = {};
            ofn.lStructSize = sizeof(ofn);
            ofn.hwndOwner = m_dlg;
            ofn.lpstrFilter = "Mainspring Time File (*.mspring)\0*.mspring\0\0";
            ofn.lpstrDefExt = "mspring";
            char fileBuf[MAX_PATH] = {};
            ofn.lpstrFile = fileBuf;
            ofn.nMaxFile = MAX_PATH;
            ofn.Flags = OFN_OVERWRITEPROMPT;

            if (GetSaveFileName(&ofn))
                m_file = fileBuf;
            else
                return;
        }

        Seconds totalTime = m_elapsedSaved + getElapsed();
        std::string saveData = std::to_string(totalTime);
        writeFile(m_file.c_str(), saveData);
        m_showSavedText = true;
        SetTimer(m_dlg, IDT_SAVEDTEXT, 3000, nullptr);
    }

    void Application::open(std::string path) {
        if (path.empty()) {
            OPENFILENAME ofn = {};
            ofn.lStructSize = sizeof(ofn);
            ofn.hwndOwner = m_dlg;
            ofn.lpstrFilter = "Mainspring Time File (*.mspring)\0*.mspring\0\0";
            ofn.lpstrDefExt = "mspring";
            char fileBuf[MAX_PATH] = {};
            ofn.lpstrFile = fileBuf;
            ofn.nMaxFile = MAX_PATH;
            ofn.Flags = OFN_FILEMUSTEXIST;

            if (GetOpenFileName(&ofn))
                m_file = fileBuf;
            else
                return;
        }
        else
            m_file = path;

        std::string openData;
        if (readFile(m_file.c_str(), openData))
            // Update status text;
            m_elapsedSaved = std::stoull(openData);
    }
}